#include <qpe/qpeapplication.h>
#include <qlineedit.h>
#include <qcheckbox.h>
#include <qlabel.h>
#include <qmessagebox.h>

#include "game.h"
#include "mainwindow.h"
#include "welcomewindow.h"
#include "handview.h"
#include "tableview.h"
#include "playerdialog.h"
#include "player.h"
#include "deck.h"
#include "suit.h"

// Global variable to hold current game info
Game* game = NULL;

//##############################################################################
Game::Game( WelcomeWindow* welcomeWindow )
{
	this->welcomeWindow = welcomeWindow;
	mainWindow = new MainWindow();
	currentPlayer = 0;
	dealer = 0;
	firstPlayer = 0;
	lastWinner = 0;
	firstTrick = true;
	spadesPlayed = false;
	deck = new Deck();
	players[0] = new Player( "Player 1", true );
	players[1] = new Player( "Player 2", false );
	players[2] = new Player( "Player 3", false );
	players[3] = new Player( "Player 4", false );
	for (int i=0; i<4; i++)
		trick[i] = NULL;
	team1Score = 0;
	team2Score = 0;
	team1Underbid = 0;
	team2Underbid = 0;
}

//##############################################################################
/**
 * Prompt user for player names and human flag
 */
void Game::setPlayers()
{
    PlayerDialog *dlg = new PlayerDialog( qApp->mainWidget(), (const char*)NULL, TRUE );

    dlg->player1Name->setText( players[0]->getName() );
    dlg->player2Name->setText( players[1]->getName() );
    dlg->player3Name->setText( players[2]->getName() );
    dlg->player4Name->setText( players[3]->getName() );

    dlg->player1Human->setChecked( players[0]->isHuman() );
    dlg->player2Human->setChecked( players[1]->isHuman() );
    dlg->player3Human->setChecked( players[2]->isHuman() );
    dlg->player4Human->setChecked( players[3]->isHuman() );

    if ( dlg->exec() ) {
	players[0]->setName( dlg->player1Name->text() );
	players[1]->setName( dlg->player2Name->text() );
	players[2]->setName( dlg->player3Name->text() );
	players[3]->setName( dlg->player4Name->text() );
	players[0]->setHuman( dlg->player1Human->isChecked() );
	players[1]->setHuman( dlg->player2Human->isChecked() );
	players[2]->setHuman( dlg->player3Human->isChecked() );
	players[3]->setHuman( dlg->player4Human->isChecked() );
    }
    else
        qApp->quit();
    delete dlg;
}

//##############################################################################
void Game::newGame()
{
	setPlayers();
	getMainWindow()->team1Names->setText(players[0]->getName()+"/"+players[2]->getName());
	getMainWindow()->team2Names->setText(players[1]->getName()+"/"+players[3]->getName());
	qApp->setMainWidget( mainWindow );
	mainWindow->showMaximized();
	dealer = 3; // Set dealer to last player so when hand starts, dealer will be first player
	startNewHand();
}

//##############################################################################
void Game::startNewHand()
{
	firstTrick = true;
	spadesPlayed = false;
	for (int i=0; i<4; i++) {
		players[i]->clearTotals();
		trick[i] = NULL;
	}
	nextDealer();
	deal();
	playedCards.clear();
	currentPlayer = dealer;
	nextPlayer(); // Set current player to player at dealer's left for first bid
	getMainWindow()->tableView->update();
	updateTeamTotals();
	players[currentPlayer]->requestBid();
}

//##############################################################################
/**
 * Change the current player to the next player.
 */
void Game::nextPlayer() 
{
	currentPlayer = (currentPlayer+1) % 4;
}

//##############################################################################
/**
 * Change the dealer to the next player.
 */
void Game::nextDealer() 
{
	dealer = (dealer+1) % 4;
}

//##############################################################################
/**
 * Request the next player's bid.  If we've reached the dealer, start playing.
 */
void Game::recordBid() 
{
	//qDebug("->Game::recordBid: player %d, bid %d, hand="+getPlayer(currentPlayer)->handToString(), currentPlayer, getPlayer(currentPlayer)->getBid());
	hideHand();
	if (getPlayer(currentPlayer)->getBid()==-1) {
		QMessageBox::information( game->getMainWindow(), "Status", "Misdeal!" );
		startNewHand();
	}
	else {
		updateTeamTotals();
		if (currentPlayer == dealer)
			playTrick(); // All bids are in, start game!
		else {
			currentPlayer = (currentPlayer+1) % 4; // Move to next player
			players[currentPlayer]->requestBid();
		}
	}
	//qDebug("Game::recordBid end");
}

//##############################################################################
/**
 * Show the specified player's hand
 */
void Game::showHand( Player* player ) {
	getMainWindow()->handView->showHand(player);
}

//##############################################################################
/**
 * Clear the hand display window so no player's hand is shown
 */
void Game::hideHand() {
	getMainWindow()->handView->hideHand();
}

//##############################################################################
/**
 * Show a message with an Ok button in the player hand window
 */
void Game::showMessage( QString message ) {
	getMainWindow()->handView->showMessage(message);
}

//##############################################################################
/**
 * Hide the message and ok button in the player hand window
 */
void Game::hideMessage() {
	getMainWindow()->handView->hideMessage();
}

//##############################################################################
Suit* Game::getLeadSuit() {
	return trick[getFirstPlayer()]->getSuit();
}

//##############################################################################
void Game::cardPlayed(Card* card) {
	//qDebug("->Game::cardPlayed(%p) currentPlayer=%d", card, currentPlayer);
	trick[currentPlayer] = card;
	getMainWindow()->tableView->update();
	currentPlayer = (currentPlayer+1) % 4;
	bool trickDone = true;
	for(int i = 0; i < 4; i++) {
		if(trick[i] == NULL)
			trickDone = false;
	}
	//Remember top 3 cards of each suit
	if(card->getValue() > 10)
		playedCards.append(card);

	if(trickDone) {
		lastWinner = calculateWinner();
		waitForUserToAcknowledgeWinner();
	}
	else
		playCard();
	//qDebug("<-Game::cardPlayed(%p) currentPlayer=%d", card, currentPlayer);
}

//##############################################################################
void Game::waitForUserToAcknowledgeWinner()
{
	// Show main window with first player at the bottom
	currentPlayer = 0;
	getMainWindow()->update();

	// Update trick counters
	players[lastWinner]->setTricks( players[lastWinner]->getTricks()+1 );
	updateTeamTotals();

	// Tell user who won the trick and wait for response
	showMessage( players[lastWinner]->getName()+" wins trick!" );
	connect( (QObject*)game->getMainWindow()->handView->button, SIGNAL( clicked() ), (QObject*)this, SLOT( trickFinished() ) );
}

//##############################################################################
void Game::trickFinished()
{
	hideMessage();
	if(players[0]->getCardCount() == 0)
		handFinished();
	else
		playTrick();
}

//##############################################################################
void Game::handFinished()
{
	// Update scores
	team1Score += calculateScore( getPlayer(0), getPlayer(2) );
	team2Score += calculateScore( getPlayer(1), getPlayer(3) );

	// Update underbid count
	team1Underbid = calculateUnderbid( getPlayer(0), getPlayer(2) );
	team2Underbid = calculateUnderbid( getPlayer(1), getPlayer(3) );

	updateTeamTotals();

	// Deduct 100 from score if underbid reached 10
	if (team1Underbid >= 10) {
		team1Score -= 100;
		QMessageBox::information( game->getMainWindow(), getPlayer(0)->getName()+"/"+getPlayer(2)->getName(), "Underbid 10, lose 100 pts." );
		team1Underbid -= 10;
		updateTeamTotals();
	}
	if (team2Underbid >= 10) {
		team2Score -= 100;
		QMessageBox::information( game->getMainWindow(), getPlayer(1)->getName()+"/"+getPlayer(3)->getName(), "Underbid 10, lose 100 pts." );
		team2Underbid -= 10;
		updateTeamTotals();
	}

	// Clear trick cards off table	
	for (int i=0; i<4; i++) {
		players[i]->clearTotals();
		trick[i] = NULL;
	}
	getMainWindow()->tableView->update();
	
	// Reset underbid if it reached 10
	if (team1Underbid >= 10)
		team1Underbid -= 10;
	if (team2Underbid >= 10)
		team2Underbid -= 10;

	if ((team1Score >= 500 || team2Score >= 500) && team1Score != team2Score)
		gameFinished();
	else {
		if (team1Score >= 500 && team2Score == team1Score)
			QMessageBox::information( game->getMainWindow(), "Game Tied", "Play a tie-breaker!" );
		else
			QMessageBox::information( game->getMainWindow(), "Status", "Hand finished." );
		startNewHand();
	}
}

//##############################################################################
/**
 * Show winner and go back to the welcome window
 */
void Game::gameFinished()
{
	if (team1Score > team2Score)
		QMessageBox::information( game->getMainWindow(), "Game Over", getPlayer(0)->getName()+"/"+getPlayer(2)->getName()+" wins!" );
	else if (team2Score > team1Score)
		QMessageBox::information( game->getMainWindow(), "Game Over", getPlayer(1)->getName()+"/"+getPlayer(3)->getName()+" wins!" );
	qApp->setMainWidget( (QWidget*)welcomeWindow );
	welcomeWindow->show();
}

//##############################################################################
/**
 * Update team totals
 */
void Game::updateTeamTotals()
{
	getMainWindow()->team1Tricks->setText(QString::number(players[0]->getTricks())+"/"+QString::number(players[2]->getTricks()));
	getMainWindow()->team2Tricks->setText(QString::number(players[1]->getTricks())+"/"+QString::number(players[3]->getTricks()));
	QString bidPlayer0 = QString::number(players[0]->getBid());
	getMainWindow()->team1Bid->setText(getBidString(0)+"/"+getBidString(2));
	getMainWindow()->team2Bid->setText(getBidString(1)+"/"+getBidString(3));
	getMainWindow()->team1Score->setText(QString::number(team1Score));
	getMainWindow()->team2Score->setText(QString::number(team2Score));
	getMainWindow()->team1Underbid->setText(QString::number(team1Underbid));
	getMainWindow()->team2Underbid->setText(QString::number(team2Underbid));
}

//##############################################################################
/**
 * Return string to represent player's bid.  " " means player has not bid yet,
 * "N" means player has bid nil, a number means the player made a normal bid.
 */
QString Game::getBidString(int playerIndex) {
	QString bidString = " "; // Assume player has not bid yet (bid == -1)
	int bid = players[playerIndex]->getBid();
	if (bid == 0)
		bidString = "N";
	else if (bid > 0)
		bidString = QString::number(bid);
	return bidString;
}

//##############################################################################
/**
 * Calculate the score for a team at the end of a hand.
 * If either partner bid nil, calculate the player scores separately,
 * otherwise, calculate the score based on the team total bid and tricks.
 */
int Game::calculateScore( Player* playerA, Player* playerB )
{
	int score = 0;
	int teamBid = playerA->getBid() + playerB->getBid();
	int teamTricks = playerA->getTricks() + playerB->getTricks();
	score += calculateBidScore( teamBid, teamTricks );
	if (playerA->getBid()==0 && playerB->getBid()==0)
		score += calculateBidScore( teamBid, teamTricks ) + teamTricks;
	else if (playerA->getBid()==0)
		score += calculateBidScore( playerA->getBid(), playerA->getTricks() );
	else if (playerB->getBid()==0)
		score += calculateBidScore( playerB->getBid(), playerB->getTricks() );
	return score;
}

//##############################################################################
/**
 * Calculate the score of a player whose partner bid nil
 */
int Game::calculateBidScore( int bid, int tricks )
{
	int score = 0;
	if (bid == 0 && tricks==0)
		score += 100;
	if (bid == 0 && tricks>0)
		score -= 100;
	else if (bid > tricks)
		score -= bid * 10;
	else
		score += (bid * 10) + (tricks - bid);
	return score;
}

//##############################################################################
/**
 * Calculate the number of tricks taken above the number of tricks bid.
 */
int Game::calculateUnderbid( Player* playerA, Player* playerB )
{
	int overbid = 0;
	int teamBid = playerA->getBid() + playerB->getBid();
	int teamTricks = playerA->getTricks() + playerB->getTricks();
	if (teamTricks > teamBid)
		overbid = teamTricks - teamBid;
	return overbid;
}

//##############################################################################
bool Game::cardHasBeenPlayed(Suit* suit, int value) {
	for(uint i = 0; i < playedCards.count(); i++) {
		Card* card = playedCards.at(i);
		if(card->getSuit() == suit && card->getValue() == value)
			return true;
	}
	return false;
}

//##############################################################################
void Game::playTrick() {
	//qDebug("Game::playTrick start");
	if(firstTrick) {
		 currentPlayer = (dealer+1) % 4;
		 firstTrick = false;
	}
	else
		currentPlayer = lastWinner;
	firstPlayer = currentPlayer;
	for(int i = 0; i < 4; i++)
		trick[i] = NULL;
	getMainWindow()->tableView->update();
	playCard();
	//qDebug("Game::playTrick end");
}

//##############################################################################
void Game::playCard() {
	//qDebug("->Game::playCard");
	if(players[currentPlayer]->isHuman())
		players[currentPlayer]->waitForReadyToSelectCard();
	else
		cardPlayed(players[currentPlayer]->calculateCard(currentPlayer==firstPlayer));
	//qDebug("<-Game::playCard");
}

//public Card[] getPlayedCards() {
//	return (Card[])playedCards.toArray();
//}

//##############################################################################
/**
 * Calculates the index of the player with the highest card in the trick.
 * Returns 0 if there are no cards in the trick.
 */
int Game::calculateWinner() {
	//qDebug("->Game::calculateWinner");
	int winner = 0;
	bool spadesFound = false;
	for(int i = 0; !spadesFound && i < 4; i++) {
		if(trick[i]!=NULL && trick[i]->getSuit() == SPADES)
			spadesFound = true;
	}
	Suit* leadSuit = getLeadSuit();
	if(spadesFound) {
		leadSuit = SPADES;
		spadesPlayed = true;
	}

	Card* highCard = NULL;
	for(int i = 0; i < 4; i++) {
		Card* card = trick[i];
		if (card!=NULL && card->getSuit()==leadSuit && 
		    (highCard==NULL || card->getValue()>highCard->getValue())) 
		{
			highCard = card;
			winner = i;
		}
	}
	//qDebug("<-Game::calculateWinner");
	return winner;
}

//##############################################################################
Card* Game::getHighestTrickCard() {
	return trick[calculateWinner()];
}

//##############################################################################
void Game::deal() {
	for(int i = 0; i < 4; i++) {
		players[i]->throwAwayCards();
	}
	deck->shuffle();
	for(int i = 0; i < 13; i++) {
		for(int j = 0; j < 4; j++) {
			players[j]->addCard( deck->dealCard() );
		}
	}
}

//##############################################################################
bool Game::isValidCardToPlay(int cardIndex) 
{
	//qDebug("->Game::isValidCardToPlay(%d)", cardIndex);
	bool validCard = false;
	Player* player = players[currentPlayer];
	if(getCurrentPlayer() == getFirstPlayer()) {
		// Check for valid lead card
		if(player->getCard(cardIndex)->getSuit() != SPADES)
			validCard = true; // Any non-spade lead card is valid
		else if(getSpadesPlayed())
			validCard = true; // If spades has been played, any lead card is valid
		else if (player->hasOnlySpades())
			validCard = player->hasOnlySpades(); // Player can lead spades if that's all he has
		else
			QMessageBox::warning( getMainWindow(), "Select another card", "Spades can't lead, none played yet!" );
	}
	else {
		// Check for valid card when player is not leading
		if(getLeadSuit() == player->getCard(cardIndex)->getSuit())
			validCard = true; // If player followed suit, card is valid
		else if (!player->hasSuit( getLeadSuit() ))
			validCard = true; // If player has none of the lead suit, any card is valid
		else
			QMessageBox::warning( getMainWindow(), "Select another card", "Please play a card in the lead suit." );
	}
	//qDebug("<-Game::isValidCardToPlay(%d)", cardIndex);
	return validCard;
}

//##############################################################################
/**
 * If current player's partner bid more than 0 and less than 4, set minimum bid.
 */
int Game::getMinimumBid()
{
	int minimumBid = 0;
	int partnerIndex = (currentPlayer+2)%4;
	int partnerBid = getPlayer(partnerIndex)->getBid();
	if (partnerBid > 0 && partnerBid < 4)
		minimumBid = 4-partnerBid;
	return minimumBid;
}

//##############################################################################
/**
 * Get the index of the specified player.
 */
int Game::getPlayerIndex( Player *player ) {
	int playerIndex = -1;
	for (int i=0; playerIndex==-1 && i<4; i++)
		if (players[i] == player)
			playerIndex = i;
	return playerIndex;
}

//##############################################################################
/**
 * Get the partner of the specified player.
 */
Player* Game::getPartner( Player* player ) {
	int index = getPlayerIndex( player );
	return players[(index+2)%4];
}

//##############################################################################
/**
 * Get the player to the left of the specified player
 */
Player* Game::getPlayerOnLeft( Player* player ) {
	int index = getPlayerIndex( player );
	return players[(index+1)%4];
}

//##############################################################################
/**
 * Get the player to the right of the specified player
 */
Player* Game::getPlayerOnRight( Player* player ) {
	int index = getPlayerIndex( player );
	return players[(index+3)%4];
}

//##############################################################################
/**
 * Get the card played in the current trick by the specified player.
 * Card will be NULL if player hasn't played yet.
 */
Card* Game::getTrickCard( Player* player ) {
	int index = getPlayerIndex( player );
	return trick[index];
}
