#include "player.h"
#include "game.h"
#include "mainwindow.h"
#include "biddialog.h"
#include "handview.h"
#include "suit.h"

#include <qpe/qpeapplication.h>
#include <qspinbox.h>
#include <qpushbutton.h>
#include <qmessagebox.h>

//##############################################################################
Player::Player( QString name, bool human ) {
	this->name = name;
	this->human = human;
	clearTotals();
}

//##############################################################################
void Player::clearTotals() 
{
	bid = -1; // -1 means no bid yet, 0 means nil bid, >0 means normal bid
	tricks = 0;
}

//##############################################################################
/**
 * Add a card to the player's hand sorted by suit and value
 */
void Player::addCard(Card* card) 
{
	//qDebug("Player::addCard( %d %s%s", card->getValue(), card->getValueName()->latin1(), card->getSuit()->getName().latin1() );
	int higherCard = -1;
	int count = cards.count();
	for(int i = 0; higherCard == -1 && i < count; i++) {
		Card* playerCard = cards.at(i);
		if(playerCard->getSuit() > card->getSuit() ||
		   (playerCard->getSuit() == card->getSuit() && playerCard->getValue() > card->getValue()))
			higherCard = i;
	}
	if(higherCard == -1)
		cards.append( card );
	else
		cards.insert( higherCard, card );
	//qDebug("Player::addCard end" );
}

//##############################################################################
void Player::throwAwayCards() {
	cards.clear();
}

//##############################################################################
int Player::getCardCount() {
	return cards.count();
}

//##############################################################################
Card* Player::getCard(int index) {
	return cards.at(index);
}

//##############################################################################
void Player::calculateBid() {
	bid = 0;
	for(int i=0; i<getCardCount(); i++) {
		Card* card = getCard(i);
		int cardsInSuit = getTotalCardsInSuit( card->getSuit() );
		if (card->getValue()==14)
			bid++; // Bid all aces
		if (card->getValue()==13 && cardsInSuit>1)
			bid++; // Bid kings unless that the only card we have in the suit
		if (card->getValue()==12 && cardsInSuit==3 && haveCard(13,card->getSuit()) && haveCard(14,card->getSuit()))
			bid++; // Bid queens if you only have the ace and king in the suit
	}
	int shortSuitCount = 0;
	shortSuitCount += getShortSuitCount( CLUBS );
	shortSuitCount += getShortSuitCount( DIAMONDS );
	shortSuitCount += getShortSuitCount( HEARTS );
	int spades = getTotalCardsInSuit( SPADES );
	if (shortSuitCount > spades)
		shortSuitCount = spades; // Don't count short suits if no spades to cover them
	bid += shortSuitCount;
	if (bid == 0)
		bid = 1; // We must bid at least one unless we plan to bid nil
	if (bid < 2 && shouldPlayNil())
		bid = 0;
	else if (bid < 2 && misdealAllowed())
		bid = -1; // Call for misdeal by setting bid to -1
	else if (bid < game->getMinimumBid())
		bid = game->getMinimumBid();
}

//##############################################################################
/**
 * Return a string with all the cards in the player's hand
 */
QString Player::handToString()
{
	QString string;
	for(int i=0; i<getCardCount(); i++) {
		Card* card = getCard(i);
		if (string.length()>0)
			string.append(" ");
		string.append( *card->getValueName() );
		string.append( card->getSuit()->getName() );
	}
	return string;
}

//##############################################################################
/**
 * Checks the specified suit to see if player has less than 3 cards and, if so,
 * returns 3 minus the number of cards in the suit.
 */
int Player::getShortSuitCount( Suit* suit )
{
	int shortSuitCount = 0;
	int suitCount = getTotalCardsInSuit( suit );
	if (suitCount < 3)
		shortSuitCount = (3-suitCount);
	return shortSuitCount;
}

//##############################################################################
/**
 * Return true if we have the specified card in the specified suit.
 */
bool Player::haveCard(int value, Suit* suit) 
{
	bool foundIt = false;
	for(int i=0; i<getCardCount(); i++) {
		Card* card = getCard(i);
		if(card->getSuit()==suit && card->getValue()==value)
			foundIt = true;
	}
	return foundIt;
}

//##############################################################################
/**
 * Get the lowest card in the players had in the specified suit.
 * Returns NULL if player has no cards in the suit.
 */
int Player::getTotalCardsInSuit( Suit* suit ) 
{
	int count = 0;
	for(int i=0; i<getCardCount(); i++) {
		if (getCard(i)->getSuit()==suit)
			count++;
	}
	return count;
}

//##############################################################################
void Player::waitForReadyToSelectCard() 
{
	//qDebug("->Player::waitForReadyToSelectCard");
	if (getCardCount()==13)
		warnOfNilBids();
	game->showMessage( "Get ready to play card..." );
	connect( (QObject*)game->getMainWindow()->handView->button, SIGNAL( clicked() ), (QObject*)this, SLOT( waitForCardToBeSelected() ) );
	//qDebug("<-Player::waitForReadyToSelectCard");
}

//##############################################################################
/**
 * Called when user clicks button telling system he is ready to select a card to play
 */
void Player::waitForCardToBeSelected() 
{
	//qDebug("->Player::waitForCardToBeSelected");
	game->hideMessage();
	game->showHand(this);
	connect( (QObject*)game->getMainWindow()->handView, SIGNAL(cardSelected(int)), (QObject*)this, SLOT(playSelectedCard(int)) );
	//qDebug("<-Player::waitForCardToBeSelected");
}


//##############################################################################
void Player::playSelectedCard(int i) 
{
	//qDebug("->Player::playSelectedCard(%d)", i);
	if(game->isValidCardToPlay(i)) {
		// Stop waiting for user to select card
		game->hideHand();
		game->cardPlayed(cards.take(i));
 	}
	//qDebug("<-Player::playSelectedCard(%d)", i);
}

//##############################################################################
/**
 * Calculate which card should be played in the current trick based on who is
 * playing NIL if anyone.
 */
Card* Player::calculateCard(bool isLeadPlayer) {
	Card* card = NULL;
	Card* partnerCard = game->getTrickCard( game->getPartner(this) );
	if (getBid()==0 && getTricks()==0)
		card = calculateCardWithMeNil( isLeadPlayer ); // I'm playing nil and not busted yet
	else if (game->getPartner(this)->getBid()==0 && game->getPartner(this)->getTricks()==0)
		card = calculateCardWithPartnerNil( isLeadPlayer, partnerCard ); // Partner playing nil and not busted yet
	else if (game->getPlayerOnRight(this)->getBid()==0 && game->getPlayerOnRight(this)->getTricks()==0)
		card = calculateCardWithOpponentNil( isLeadPlayer ); // Right player playing nil and not busted yet
	else if (game->getPlayerOnLeft(this)->getBid()==0 && game->getPlayerOnLeft(this)->getTricks()==0)
		card = calculateCardWithOpponentNil( isLeadPlayer ); // Left player playing nil and not busted yet
	else
		card = calculateCardWithNoNil( isLeadPlayer, partnerCard ); // Nobody playing nil (or nil already busted)
	cards.remove( card );
	return card;
}

//##############################################################################
/**
 * Calculate which card to play in the current trick with nobody playing nil.
 */
Card* Player::calculateCardWithNoNil(bool isLeadPlayer, Card* partnerCard) {
	Card* card = NULL;
	if (isLeadPlayer) {
		card = findHighCardInSuit(); // Lead highest card in suit if we have one
		if (card == NULL)
			card = findLowestCard( false ); // Lead lowest non-spade if we have one
		if (card == NULL)
			card = findLowestCard( true ); // Lead lowest card we have
	}
	else {
		// Find best card to play when this is not the lead card
		Suit* leadSuit = game->getLeadSuit();
		Card* highestTrickCard = game->getHighestTrickCard();
		card = getHighestCardInSuit( leadSuit );
		if (card!=NULL) {
			// We have cards in the lead suit
			// If lead suit has been trumped throw a low card
			if (leadSuit != SPADES && highestTrickCard->getSuit() == SPADES)
				card = getLowestCardInSuit( leadSuit );

			// If our partner played the highest card in the suit, throw a low card
			else if (partnerCard!=NULL && isHighCard(partnerCard))
				card = getLowestCardInSuit( leadSuit );

			// If our partner played the king and we have the ace, throw a low card
			else if (partnerCard!=NULL && partnerCard->getValue()==13 && card->getValue()==14)
				card = getLowestCardInSuit( leadSuit );

			// If our partner played the queen and we have the king and ace, throw a low card
			else if (partnerCard!=NULL && partnerCard->getValue()==12 && card->getValue()==14 && haveCard(13,leadSuit))
				card = getLowestCardInSuit( leadSuit );

			// If we have a king or queen but another player has a higher card, throw a low card
			else if ((card->getValue()==13 || card->getValue()==12) && !isHighCard(card))
				card = getLowestCardInSuit( leadSuit );

			// If we can't beat the cards already played in the trick throw a low card
			else if (card->getValue() < highestTrickCard->getValue())
				card = getLowestCardInSuit( leadSuit );
		}
		else {
			// We have no cards in lead suit, play spade if our partner hasn't already played a spade.
			if (leadSuit != SPADES) {
				// If our partner already played a spade and is winning, play our lowest non-spade
				if (partnerCard!=NULL && partnerCard->getSuit()==SPADES && partnerCard==highestTrickCard)
					card = findLowestCard( false );

				// Else, if our partner played highest card in lead suit and no spades, play our lowest non-spade
				else if (partnerCard!=NULL && partnerCard->getSuit()==leadSuit && isHighCard(partnerCard))
					card = findLowestCard( false );

				// Else, if no spades have been played, throw our lowest spade
				else if (highestTrickCard->getSuit()!=SPADES)
					card = getLowestCardInSuit( SPADES );

				// Else, throw our lowest spade that will win
				else if (highestTrickCard->getSuit()!=SPADES)
					card = getLowestWinningSpade();
			}
			if (card == NULL)
				card = findLowestCard( true );
		}
	}
	return card;
}

//##############################################################################
/**
 * Calculate which card to play in the current trick when playing nil.
 * If leading, throw lowest card, otherwise throw highest card
 * that does not beat the cards already played.
 */
Card* Player::calculateCardWithMeNil(bool isLeadPlayer) {
	Card* card = NULL;
	if (isLeadPlayer) {
		card = findLowestCard( false ); // Lead lowest non-spade if we have one
		if (card == NULL)
			card = findLowestCard( true ); // Lead lowest card we have
	}
	else {
		// Find best card to play when this is not the lead card
		Suit* leadSuit = game->getLeadSuit();
		Card* highestTrickCard = game->getHighestTrickCard();
		card = getHighestCardInSuit( leadSuit );
		if (card!=NULL) {
			// We have cards in the lead suit
			// If lead suit has been trumped throw our highest card in lead suit
			if (leadSuit != SPADES && highestTrickCard->getSuit() == SPADES)
				; // Do nothing, we already have the right card to play

			// If highest card in trick beats our highest card in the suit, play it
			else if (highestTrickCard->getValue() > card->getValue())
				; // Do nothing, we already have the right card to play

			// If trick has not been trumped, play highest card that will lose the trick
			else
				card = getHighestLosingCardInSuit( leadSuit, highestTrickCard );

			// If we don't have any cards that will lose to the cards in the trick, play lowest card in suit
			if (card == NULL)
				card = getLowestCardInSuit( leadSuit );
		}
		else {
			// We have no cards in lead suit, play highest card that will still lose.
			if (highestTrickCard->getSuit()==SPADES)
				card = getHighestLosingCardInSuit( SPADES, highestTrickCard );

			// If we couldn't play a spades, play highest card in any other suit.
			if (card == NULL)
				card = findHighestCard( false );

			// If we only have spades, play the lowest one
			if (card == NULL)
				card = findLowestCard( true );
		}
	}
	return card;
}

//##############################################################################
/**
 * Calculate which card to play in the current trick with nobody playing nil.
 * TODO If leading and partner is short suited, lead the suit he does not have.
 * If partner hasn't played, play highest possible card unless opponent played high card.
 * If partner has played, play higher card unless opponent played higher card.
 */
Card* Player::calculateCardWithPartnerNil(bool isLeadPlayer, Card* partnerCard) {
	Card* card = NULL;
	if (isLeadPlayer) {
		card = findHighCardInSuit(); // Lead highest card in suit if we have one
		if (card == NULL)
			card = findHighestCard( false ); // Lead highest non-spade card we have
		if (card == NULL)
			card = findHighestCard( true ); // Lead highest card we have
	}
	else {
		// Find best card to play when not leading
		Suit* leadSuit = game->getLeadSuit();
		Card* highestTrickCard = game->getHighestTrickCard();
		card = getHighestCardInSuit( leadSuit );
		if (card!=NULL) {
			// We have cards in the lead suit

			// If lead suit has been trumped throw a low card
			if (leadSuit != SPADES && highestTrickCard->getSuit() == SPADES)
				card = getLowestCardInSuit( leadSuit );

			// If the highest card in the suit has already been played, play a low card
			else if (isHighCard( highestTrickCard ))
				card = getLowestCardInSuit( leadSuit );

			// If our high card is 1 higher than the highest trick card or less, play a low card
			else if (card->getValue()<=highestTrickCard->getValue()+1)
				card = getLowestCardInSuit( leadSuit );

			// Partner has already played card, beat it if necessary and we can, otherwise, play low
			else if (partnerCard != NULL) {
				// If our partner trumped the trick, play the lowest card we have in the lead suit
				if (leadSuit != SPADES && partnerCard->getSuit()==SPADES)
					card = getLowestCardInSuit( leadSuit );

				// If our partner played but an opponenet beat his card, play the lowest card we have in the lead suit
				else if (partnerCard!=highestTrickCard)
					card = getLowestCardInSuit( leadSuit );

				// If our partner played the highest card in the trick, try to beat it
				else if (partnerCard==highestTrickCard) {
					card = getLowestWinningCardInSuit( leadSuit, partnerCard );
					if (card == NULL)
						card = getLowestCardInSuit( leadSuit );
				}
			}
		}
		else {
			// We have no cards in lead suit, play spade if our partner hasn't already played a spade.
			if (leadSuit != SPADES) {
				// If our partner has played, try to beat it if necessary and we can
				if (partnerCard!=NULL) {
					// If partner is not winning, play our lowest non-spade
					if (partnerCard!=highestTrickCard)
						card = findLowestCard( false );

					// If our partner is winning with a spade, try to beat it
					else if (partnerCard->getSuit()==SPADES)
						card = getLowestWinningCardInSuit( SPADES, partnerCard );

					// If our partner is winning with a non-spade, play our lowest spade
					else if (partnerCard!=NULL)
						card = getLowestCardInSuit( SPADES );
				}
				// Else, if our partner hasn't played but the trick has been trumped, play a low card
				else if (leadSuit!=SPADES && highestTrickCard->getSuit()==SPADES)
					card = findLowestCard( false );

				// Else, if no spades have been played, play our lowest spade to cover our partner
				else if (highestTrickCard->getSuit()!=SPADES)
					card = getLowestCardInSuit( SPADES );
			}
			if (card == NULL)
				card = findLowestCard( false );
			if (card == NULL)
				card = findLowestCard( true );
		}
	}
	return card;
}

//##############################################################################
/**
 * Calculate which card to play in the current trick when opponent playing nil.
 * If opponent is playing nil, play low cards to try to bust nil.
 */
Card* Player::calculateCardWithOpponentNil(bool isLeadPlayer) {
	Card* card = NULL;
	if (isLeadPlayer) {
		card = findLowestCard( false ); // Lead lowest non-spade if we have one
		if (card == NULL)
			card = findLowestCard( true ); // Lead lowest card we have
	}
	else {
		// Figure out which opponent is playing nil and get his trick card
		Player* nilPlayer = game->getPlayerOnLeft(this);
		if (nilPlayer->getBid()>0 || nilPlayer->getTricks()>0)
			nilPlayer = game->getPlayerOnRight(this);
		Card* nilPlayersTrickCard = game->getTrickCard( nilPlayer );

		// Find best card to play when this is not the lead card
		Suit* leadSuit = game->getLeadSuit();
		Card* highestTrickCard = game->getHighestTrickCard();
		card = getHighestCardInSuit( leadSuit );
		if (card!=NULL) {
			// We have cards in the lead suit

			// If lead suit has been trumped throw our highest card in lead suit
			if (leadSuit != SPADES && highestTrickCard->getSuit() == SPADES)
				; // Do nothing, we already have the right card to play

			// If highest card in trick beats our highest card in the suit, play it
			else if (highestTrickCard->getValue() > card->getValue())
				; // Do nothing, we already have the right card to play

			// If nil player has already lost trick, try to take it with highest card we have in suit
			else if (nilPlayersTrickCard != NULL && nilPlayersTrickCard != highestTrickCard)
				; // Do nothing, we already have the right card to play

			// If trick has not been trumped, play highest card that will lose the trick
			else
				card = getHighestLosingCardInSuit( leadSuit, highestTrickCard );

			// If we don't have any cards that will lose to the cards in the trick, play highest card in suit
			if (card == NULL)
				card = getHighestCardInSuit( leadSuit );
		}
		else {
			// We have no cards in lead suit

			// If nilPlayer has already lost trick, try to take it with highest card we have
			if (nilPlayersTrickCard != NULL && nilPlayersTrickCard != highestTrickCard)
				card = findHighestCard( true );

			// If nilPlayer is winning trick, play a card that will lose to his card if possible
			else if (highestTrickCard->getSuit()==SPADES)
				card = getHighestLosingCardInSuit( SPADES, highestTrickCard );

			// If we couldn't play a spades, play highest card in any other suit.
			if (card == NULL)
				card = findHighestCard( false );

			// If we only have spades, play the highest one
			if (card == NULL)
				card = findHighestCard( true );
		}
	}
	return card;
}

//##############################################################################
/**
 * Get highest card in the suit that will lose to the highest card in the trick.
 * Returns NULL if player has no cards that will lost to the highest card.
 */
Card* Player::getHighestLosingCardInSuit( Suit* suit, Card* highestTrickCard ) {
	Card* card = NULL;
	for (int i=0; i<getCardCount(); i++) {
		Card* tempCard = getCard(i);
		if (tempCard->getSuit()==suit) {
			if ((card==NULL || tempCard->getValue()>card->getValue()) &&
			    tempCard->getValue()<highestTrickCard->getValue())
				card = tempCard;
		}
	}
	return card;
}

//##############################################################################
/**
 * Get lowest card in the suit that will beat to the highest card in the trick.
 * Returns NULL if player has no cards that will beat the highest card.
 */
Card* Player::getLowestWinningCardInSuit( Suit* suit, Card* highestTrickCard ) {
	Card* card = NULL;
	for (int i=0; i<getCardCount(); i++) {
		Card* tempCard = getCard(i);
		if (tempCard->getSuit()==suit) {
			if ((card==NULL || tempCard->getValue()<card->getValue()) &&
			    tempCard->getValue()>highestTrickCard->getValue())
				card = tempCard;
		}
	}
	return card;
}

//##############################################################################
/**
 * Get the highest card in the players hand in the specified suit.
 * Returns NULL if player has no cards in the suit.
 */
Card* Player::getHighestCardInSuit( Suit* suit ) 
{
	Card* highestCard = NULL;
	for(int i=0; i<getCardCount(); i++) {
		Card* card = getCard(i);
		if (card->getSuit()==suit && (highestCard==NULL || card->getValue() > highestCard->getValue()))
			highestCard = card;
	}
	return highestCard;
}

//##############################################################################
/**
 * Get the lowest card in the players had in the specified suit.
 * Returns NULL if player has no cards in the suit.
 */
Card* Player::getLowestCardInSuit( Suit* suit ) 
{
	Card* lowestCard = NULL;
	for(int i=0; i<getCardCount(); i++) {
		Card* card = getCard(i);
		if (card->getSuit()==suit && (lowestCard==NULL || card->getValue() < lowestCard->getValue()))
			lowestCard = card;
	}
	return lowestCard;
}

//##############################################################################
/**
 * Get the lowest card in the players had in the specified suit that will beat
 * the cards already played in the trick.
 * Returns NULL if player has no spades that can be the highest card in the trick.
 */
Card* Player::getLowestWinningSpade() 
{
	Card* highestTrickCard = game->getHighestTrickCard();
	Card* lowestCard = NULL;
	for(int i=0; i<getCardCount(); i++) {
		Card* card = getCard(i);
		if (card->getSuit()==SPADES && 
		    (highestTrickCard->getSuit()!=SPADES || card->getValue() > highestTrickCard->getValue()) &&
		    (lowestCard==NULL || card->getValue() < lowestCard->getValue()))
			lowestCard = card;
	}
	return lowestCard;
}

//##############################################################################
/**
 * Checks player's hand to see if he has the highest unplayed card in any suit.
 * Returns NULL if none of the cards in the hand are the highest unplayed cards
 * in their suit.
 */
Card* Player::findHighCardInSuit()
{
	Card* highCard = NULL;
	for(int i=0; highCard==NULL && i<getCardCount(); i++) {
		Card *card = getCard(i);
		if((game->getSpadesPlayed() || card->getSuit()!=SPADES) && isHighCard( card ))
			highCard = card;
	}
	return highCard;
}

//##############################################################################
/**
 * Checks to see if card is the highest unplayed card in its suit.
 */
bool Player::isHighCard( Card* card )
{
	bool isHighCard = true;
	for(int i=card->getValue()+1; isHighCard && i<=14; i++) {
		isHighCard = game->cardHasBeenPlayed( card->getSuit(), i );
	}
	return isHighCard;
}

//##############################################################################
/**
 * Finds the highest card in the players hand excluding spades unless the
 * includeSpades flag is true.
 * Returns NULL if includeSpades is false and hand contains only spades.
 */
Card* Player::findHighestCard( bool includeSpades )
{
	Card* highCard = NULL;
	for(int i=0; i<getCardCount(); i++) {
		Card *card = getCard(i);
		if(highCard==NULL && (includeSpades || card->getSuit()!=SPADES))
			highCard = card;
		else if (highCard!=NULL && (includeSpades || card->getSuit()!=SPADES) && card->getValue()>highCard->getValue())
			highCard = card;
	}
	return highCard;
}

//##############################################################################
/**
 * Finds the lowest card in the players hand excluding spades unless the
 * includeSpades flag is true.
 * Returns NULL if includeSpades is false and hand contains only spades.
 */
Card* Player::findLowestCard( bool includeSpades )
{
	Card* lowCard = NULL;
	for(int i=0; i<getCardCount(); i++) {
		Card *card = getCard(i);
		if(lowCard==NULL && (includeSpades || card->getSuit()!=SPADES))
			lowCard = card;
		else if(lowCard!=NULL && (includeSpades || card->getSuit()!=SPADES) && card->getValue()<lowCard->getValue())
			lowCard = card;
	}
	return lowCard;
}

//##############################################################################
int Player::haveCard(int value, bool includeSpades) {
	for(int j = 0; j < getCardCount(); j++) {
		Card* card = getCard(j);
		if((includeSpades || card->getSuit() != SPADES) && card->getValue() == value) {
			return j;
		}
	}
	return -1;
}

//##############################################################################
bool Player::hasOnlySpades() {
	bool allSpades = true;
	for(int j = 0; allSpades && j < getCardCount(); j++) {
		allSpades = getCard(j)->getSuit() == SPADES;
	}
	return allSpades;
}

//##############################################################################
bool Player::hasSuit( Suit* suit ) {
	bool foundSuit = false;
	for(int j = 0; !foundSuit && j < getCardCount(); j++) {
		foundSuit = getCard(j)->getSuit() == suit;
	}
	return foundSuit;
}

//##############################################################################
void Player::requestBid() {
	//qDebug("requestBid for "+getName()+" start");
	
	if(human) {
		warnOfNilBids();
		game->showMessage( "Get ready to bid..." );
		connect( (QObject*)game->getMainWindow()->handView->button, SIGNAL(clicked()), (QObject*)this, SLOT(promptForBid()));
	}
	else {
		calculateBid();
		game->recordBid();
	}
	//qDebug("requestBid for "+getName()+" end");
}

//##############################################################################
void Player::promptForBid() {
	//qDebug("Player::promptForBid start");
	game->hideMessage();
	game->showHand(this);
	int minimumBid = game->getMinimumBid();
	do {
		BidDialog *bidDialog = new BidDialog( qApp->mainWidget(), (const char*)NULL, TRUE );
		bidDialog->setCaption( getName() );
		bidDialog->cancelButton->setEnabled( misdealAllowed() );
		if ( bidDialog->exec() )
			bid = bidDialog->bid->text().toInt();
		else
			//qApp->quit();
			bid = -1; // Player clicked misdeal button
		delete bidDialog;
		if (bid > 0 && bid < minimumBid)
			QMessageBox::information( game->getMainWindow(), "Invalid Bid", "Minimum bid is nil (0) or "+QString::number(minimumBid) );
	} while (bid > 0 && bid < minimumBid);
	game->recordBid();
	//qDebug("Player::promptForBid end");
}

//##############################################################################
/**
 * Return true if player's hand has no spades or more than 5 of any other suit.
 */
bool Player::misdealAllowed() {
	return getTotalCardsInSuit( SPADES )==0 ||
	       getTotalCardsInSuit( HEARTS )>5 ||
	       getTotalCardsInSuit( DIAMONDS )>5 ||
	       getTotalCardsInSuit( CLUBS )>5;
}

//##############################################################################
/**
 * Return true if player's hand is a good nil hand.
 * A good nill hand has:
 *   - less than 3 spades
 *   - all spades lower than 10
 *   - 3 cards below 8 in any suit with a card higher than 10 OR
 *     - 2 cards below 8 in any suit with a card higher than 10 and one suit with less than 3 cards
 */
bool Player::shouldPlayNil() {
	bool nil = true;
	if (getTotalCardsInSuit( SPADES )>=3)
		nil = false;
	if (nil) {
		// Check sizes of spades
		for(int i=0; nil && i<getCardCount(); i++) {
			Card *card = getCard(i);
			if(card->getSuit()==SPADES && card->getValue()>=10)
				nil = false;
		}
	}
	if (nil) {
		int minimumLowCards = 3;
		if (getShortSuitCount( CLUBS )>0 || getShortSuitCount( DIAMONDS )>0 || getShortSuitCount( HEARTS )>0)
			minimumLowCards = 2;
		if (!canCoverHighCards( CLUBS, minimumLowCards ) ||
		    !canCoverHighCards( DIAMONDS, minimumLowCards ) ||
		    !canCoverHighCards( HEARTS, minimumLowCards ))
			nil = false;
	}
	return nil;
}

//##############################################################################
/**
 * Return true if player has enough low cards (<8) to cover any high cards in
 * the same suit when playing a nil hand.
 */
int Player::canCoverHighCards( Suit* suit, int minimumLowCards ) {
	bool canCover = true;
	int lowCards = 0;
	int highCards = 0;
	for(int i=0; i<getCardCount(); i++) {
		Card *card = getCard(i);
		if(card->getSuit()==suit) {
			if (card->getValue()<8)
				lowCards++;
			if (card->getValue()>10)
				highCards++;
		}
	}
	if (highCards > 0 && lowCards < minimumLowCards)
		canCover = false;
	return canCover;
}

//##############################################################################
/**
 * Display a message telling user about any players that bid nil.
 */
void Player::warnOfNilBids() {
	for (int i=0; i<4; i++) {
		Player* player = game->getPlayer(i);
		if (player->getBid()==0)
			QMessageBox::information( game->getMainWindow(), "Nil Bid", player->getName()+" bid NIL!" );
	}
}
